import pytest
from pytest_benchmark.fixture import BenchmarkFixture

from hdfs_native import AsyncClient, Client


@pytest.mark.benchmark
def test_benchmark_listing(client: Client, benchmark: BenchmarkFixture):
    for i in range(1000):
        client.create(f"/bench{i}").close()

    def do_work():
        for _ in client.list_status("/"):
            pass

    benchmark(do_work)


@pytest.mark.benchmark
def test_benchmark_listing_async(
    client: Client,
    async_client: AsyncClient,
    aio_benchmark,
):
    for i in range(1000):
        client.create(f"/bench{i}").close()

    async def do_work():
        async for _ in async_client.list_status("/"):
            pass

    aio_benchmark(do_work)



from collections.abc import AsyncIterator
from typing import Dict, Iterator, List, Literal, Optional

# For some reason mypy doesn't think this exists
from typing_extensions import Buffer  # type: ignore

class FileStatus:
    path: str
    length: int
    isdir: bool
    permission: int
    owner: str
    group: str
    modification_time: int
    access_time: int
    replication: Optional[int]
    blocksize: Optional[int]

class ContentSummary:
    length: int
    file_count: int
    directory_count: int
    quota: int
    space_consumed: int
    space_quota: int

AclEntryType = Literal["user", "group", "mask", "other"]
AclEntryScope = Literal["access", "default"]
FsAction = Literal["---", "--x", "-w-", "-wx", "r--", "r-x", "rw-", "rwx"]

class AclEntry:
    type: AclEntryType
    scope: AclEntryScope
    permissions: FsAction
    name: Optional[str]

    def __init__(
        self,
        type: AclEntryType,
        scope: AclEntryScope,
        permissions: FsAction,
        name: Optional[str] = None,
    ): ...

class AclStatus:
    owner: str
    group: str
    sticky: bool
    entries: List[AclEntry]
    permission: int

class WriteOptions:
    block_size: Optional[int]
    replication: Optional[int]
    permission: int
    overwrite: bool
    create_parent: bool

    def __init__(
        self,
        block_size: Optional[int] = None,
        replication: Optional[int] = None,
        permission: Optional[int] = None,
        overwrite: Optional[bool] = None,
        create_parent: Optional[bool] = None,
    ): ...

class RawFileReader:
    def file_length(self) -> int:
        """Returns the size of the file"""

    def seek(self, pos: int) -> None:
        """Sets the cursor to the given position"""

    def tell(self) -> int:
        """Returns the current cursor position in the file"""

    def read(self, len: int) -> bytes:
        """Reads `len` bytes from the file, advancing the position in the file"""

    def read_range(self, offset: int, len: int) -> bytes:
        """Read `len` bytes from the file starting at `offset`. Doesn't affect the position in the file"""

    def read_range_stream(self, offset: int, len: int) -> Iterator[bytes]:
        """
        Read `len` bytes from the file starting at `offset` as an iterator of bytes. Doesn't affect
        the position in the file.
        """

class RawFileWriter:
    def write(self, buf: Buffer) -> int:
        """Writes `buf` to the file"""

    def close(self) -> None:
        """Closes the file and saves the final metadata to the NameNode"""

class RawClient:
    def __init__(
        self,
        url: Optional[str],
        config: Optional[Dict[str, str]],
    ) -> None: ...
    def get_file_info(self, path: str) -> FileStatus: ...
    def list_status(self, path: str, recursive: bool) -> Iterator[FileStatus]: ...
    def read(self, path: str) -> RawFileReader: ...
    def create(self, path: str, write_options: WriteOptions) -> RawFileWriter: ...
    def append(self, path: str) -> RawFileWriter: ...
    def mkdirs(self, path: str, permission: int, create_parent: bool) -> None: ...
    def rename(self, src: str, dst: str, overwrite: bool) -> None: ...
    def delete(self, path: str, recursive: bool) -> bool: ...
    def set_times(self, path: str, mtime: int, atime: int) -> None: ...
    def set_owner(
        self,
        path: str,
        owner: Optional[str],
        group: Optional[str],
    ) -> None: ...
    def set_permission(self, path: str, permission: int) -> None: ...
    def set_replication(self, path: str, replication: int) -> bool: ...
    def get_content_summary(self, path: str) -> ContentSummary: ...
    def modify_acl_entries(self, path: str, entries: List[AclEntry]) -> None: ...
    def remove_acl_entries(self, path: str, entries: List[AclEntry]) -> None: ...
    def remove_default_acl(self, path: str) -> None: ...
    def remove_acl(self, path: str) -> None: ...
    def set_acl(self, path: str, entries: List[AclEntry]) -> None: ...
    def get_acl_status(self, path: str) -> AclStatus: ...

class AsyncRawFileReader:
    def file_length(self) -> int:
        """Returns the size of the file"""

    def seek(self, pos: int) -> None:
        """Sets the cursor to the given position"""

    def tell(self) -> int:
        """Returns the current cursor position in the file"""

    async def read(self, len: int) -> bytes:
        """Reads `len` bytes from the file, advancing the position in the file"""

    async def read_range(self, offset: int, len: int) -> bytes:
        """Read `len` bytes from the file starting at `offset`. Doesn't affect the position in the file"""

    def read_range_stream(self, offset: int, len: int) -> AsyncIterator[bytes]:
        """
        Read `len` bytes from the file starting at `offset` as an iterator of bytes. Doesn't affect
        the position in the file.
        """

class AsyncRawFileWriter:
    async def write(self, buf: Buffer) -> int:
        """Writes `buf` to the file"""

    async def close(self) -> None:
        """Closes the file and saves the final metadata to the NameNode"""

class AsyncRawClient:
    def __init__(
        self,
        url: Optional[str],
        config: Optional[Dict[str, str]],
    ) -> None: ...
    async def get_file_info(self, path: str) -> FileStatus: ...
    def list_status(self, path: str, recursive: bool) -> AsyncIterator[FileStatus]: ...
    async def read(self, path: str) -> AsyncRawFileReader: ...
    async def create(
        self,
        path: str,
        write_options: WriteOptions,
    ) -> AsyncRawFileWriter: ...
    async def append(self, path: str) -> AsyncRawFileWriter: ...
    async def mkdirs(self, path: str, permission: int, create_parent: bool) -> None: ...
    async def rename(self, src: str, dst: str, overwrite: bool) -> None: ...
    async def delete(self, path: str, recursive: bool) -> bool: ...
    async def set_times(self, path: str, mtime: int, atime: int) -> None: ...
    async def set_owner(
        self,
        path: str,
        owner: Optional[str],
        group: Optional[str],
    ) -> None: ...
    async def set_permission(self, path: str, permission: int) -> None: ...
    async def set_replication(self, path: str, replication: int) -> bool: ...
    async def get_content_summary(self, path: str) -> ContentSummary: ...
    async def modify_acl_entries(self, path: str, entries: List[AclEntry]) -> None: ...
    async def remove_acl_entries(self, path: str, entries: List[AclEntry]) -> None: ...
    async def remove_default_acl(self, path: str) -> None: ...
    async def remove_acl(self, path: str) -> None: ...
    async def set_acl(self, path: str, entries: List[AclEntry]) -> None: ...
    async def get_acl_status(self, path: str) -> AclStatus: ...



import io

import pytest

from hdfs_native import AclEntry, AsyncClient, Client, WriteOptions


def test_sync(client: Client):
    client.create("/testfile").close()
    file_info = client.get_file_info("/testfile")

    assert file_info.path == "/testfile"

    file_list = list(client.list_status("/", False))
    assert len(file_list) == 1
    assert file_list[0].path == "/testfile"

    client.rename("/testfile", "/testfile2", False)

    file_list = list(client.list_status("/", False))
    assert len(file_list) == 1
    assert file_list[0].path == "/testfile2"

    client.delete("/testfile2", False)

    file_list = list(client.list_status("/", False))
    assert len(file_list) == 0

    with client.create("/testfile") as file:
        data = io.BytesIO()

        for i in range(0, 32 * 1024 * 1024):
            data.write(i.to_bytes(4, "big"))

        file.write(data.getbuffer())

    with client.read("/testfile") as file:
        data = io.BytesIO(file.read())

    for i in range(0, 32 * 1024 * 1024):
        assert data.read(4) == i.to_bytes(4, "big")

    with client.append("/testfile") as file:
        data = io.BytesIO()

        for i in range(32 * 1024 * 1024, 33 * 1024 * 1024):
            data.write(i.to_bytes(4, "big"))

        file.write(data.getbuffer())

    with client.read("/testfile") as file:
        data = io.BytesIO(file.read())

        for i in range(0, 33 * 1024 * 1024):
            assert data.read(4) == i.to_bytes(4, "big")

        data = io.BytesIO()
        for chunk in file:
            data.write(chunk)

        data.seek(0)

        for i in range(0, 33 * 1024 * 1024):
            assert data.read(4) == i.to_bytes(4, "big")

    with client.read("/testfile") as file:
        # Skip first two ints
        file.seek(8)
        expected = 2
        assert file.read(4) == expected.to_bytes(4, "big")
        assert file.tell() == 12

    mtime = 1717641455
    atime = 1717641456
    client.set_times("/testfile", mtime, atime)
    file_info = client.get_file_info("/testfile")
    assert file_info.modification_time == mtime
    assert file_info.access_time == atime

    client.set_owner("/testfile", "testuser", "testgroup")
    file_info = client.get_file_info("/testfile")
    assert file_info.owner == "testuser"
    assert file_info.group == "testgroup"

    client.set_owner("/testfile", owner="testuser2")
    file_info = client.get_file_info("/testfile")
    assert file_info.owner == "testuser2"
    assert file_info.group == "testgroup"

    client.set_owner("/testfile", group="testgroup2")
    file_info = client.get_file_info("/testfile")
    assert file_info.owner == "testuser2"
    assert file_info.group == "testgroup2"

    assert file_info.permission == 0o644
    client.set_permission("/testfile", 0o600)
    file_info = client.get_file_info("/testfile")
    assert file_info.permission == 0o600

    client.set_replication("/testfile", 1)
    file_info = client.get_file_info("/testfile")
    assert file_info.replication == 1

    client.set_replication("/testfile", 2)
    file_info = client.get_file_info("/testfile")
    assert file_info.replication == 2

    content_summary = client.get_content_summary("/")
    assert content_summary.file_count == 1
    assert content_summary.directory_count == 1
    assert content_summary.length == 33 * 1024 * 1024 * 4

    client.delete("/testfile", False)


def test_write_options(client: Client):
    with client.create("/testfile") as file:
        file.write(b"abcd")

    client.create(
        "/testfile",
        WriteOptions(overwrite=True, permission=0o700, block_size=1024 * 1024),
    ).close()

    file_info = client.get_file_info("/testfile")
    assert file_info.length == 0
    assert file_info.permission == 0o700
    assert file_info.blocksize == 1024 * 1024


def test_acls(client: Client):
    client.create("/test").close()

    acl_status = client.get_acl_status("/test")
    assert len(acl_status.entries) == 0

    client.modify_acl_entries("/test", [AclEntry("user", "access", "r-x", "testuser")])
    # Should be 2 entries now, a default group entry gets added as well
    acl_status = client.get_acl_status("/test")
    assert len(acl_status.entries) == 2

    client.remove_acl("/test")
    acl_status = client.get_acl_status("/test")
    assert len(acl_status.entries) == 0

    client.delete("/test")

    client.mkdirs("/testdir")

    client.modify_acl_entries(
        "/testdir", [AclEntry("user", "default", "rwx", "testuser")]
    )
    # 4 other defaults get added automatically
    acl_status = client.get_acl_status("/testdir")
    assert len(acl_status.entries) == 5

    client.remove_default_acl("/testdir")
    acl_status = client.get_acl_status("/testdir")
    assert len(acl_status.entries) == 0


@pytest.mark.asyncio
async def test_async(async_client: AsyncClient):
    await (await async_client.create("/testfile")).close()
    file_info = await async_client.get_file_info("/testfile")

    assert file_info.path == "/testfile"

    file_list = [status async for status in async_client.list_status("/", False)]
    assert len(file_list) == 1
    assert file_list[0].path == "/testfile"

    await async_client.rename("/testfile", "/testfile2", False)

    file_list = [status async for status in async_client.list_status("/", False)]
    assert len(file_list) == 1
    assert file_list[0].path == "/testfile2"

    await async_client.delete("/testfile2", False)

    file_list = [status async for status in async_client.list_status("/", False)]
    assert len(file_list) == 0

    async with await async_client.create("/testfile") as file:
        data = io.BytesIO()

        for i in range(0, 32 * 1024 * 1024):
            data.write(i.to_bytes(4, "big"))

        await file.write(data.getbuffer())

    async with await async_client.read("/testfile") as file:
        data = io.BytesIO(await file.read())

    for i in range(0, 32 * 1024 * 1024):
        assert data.read(4) == i.to_bytes(4, "big")

    async with await async_client.append("/testfile") as file:
        data = io.BytesIO()

        for i in range(32 * 1024 * 1024, 33 * 1024 * 1024):
            data.write(i.to_bytes(4, "big"))

        await file.write(data.getbuffer())

    async with await async_client.read("/testfile") as file:
        data = io.BytesIO(await file.read())

        for i in range(0, 33 * 1024 * 1024):
            assert data.read(4) == i.to_bytes(4, "big")

        data = io.BytesIO()
        async for chunk in file:
            data.write(chunk)

        data.seek(0)

        for i in range(0, 33 * 1024 * 1024):
            assert data.read(4) == i.to_bytes(4, "big")

    async with await async_client.read("/testfile") as file:
        # Skip first two ints
        file.seek(8)
        expected = 2
        assert await file.read(4) == expected.to_bytes(4, "big")
        assert file.tell() == 12

    mtime = 1717641455
    atime = 1717641456
    await async_client.set_times("/testfile", mtime, atime)
    file_info = await async_client.get_file_info("/testfile")
    assert file_info.modification_time == mtime
    assert file_info.access_time == atime

    await async_client.set_owner("/testfile", "testuser", "testgroup")
    file_info = await async_client.get_file_info("/testfile")
    assert file_info.owner == "testuser"
    assert file_info.group == "testgroup"

    await async_client.set_owner("/testfile", owner="testuser2")
    file_info = await async_client.get_file_info("/testfile")
    assert file_info.owner == "testuser2"
    assert file_info.group == "testgroup"

    await async_client.set_owner("/testfile", group="testgroup2")
    file_info = await async_client.get_file_info("/testfile")
    assert file_info.owner == "testuser2"
    assert file_info.group == "testgroup2"

    assert file_info.permission == 0o644
    await async_client.set_permission("/testfile", 0o600)
    file_info = await async_client.get_file_info("/testfile")
    assert file_info.permission == 0o600

    await async_client.set_replication("/testfile", 1)
    file_info = await async_client.get_file_info("/testfile")
    assert file_info.replication == 1

    await async_client.set_replication("/testfile", 2)
    file_info = await async_client.get_file_info("/testfile")
    assert file_info.replication == 2

    content_summary = await async_client.get_content_summary("/")
    assert content_summary.file_count == 1
    assert content_summary.directory_count == 1
    assert content_summary.length == 33 * 1024 * 1024 * 4

    await async_client.delete("/testfile", False)



import asyncio
import os
import subprocess

import pytest
import pytest_asyncio
from pytest_benchmark.fixture import BenchmarkFixture

from hdfs_native import AsyncClient, Client


@pytest.fixture(scope="module")
def minidfs():
    child = subprocess.Popen(
        [
            "mvn",
            "-f",
            "../rust/minidfs",
            "--quiet",
            "clean",
            "compile",
            "exec:java",
        ],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        universal_newlines=True,
        encoding="utf8",
        bufsize=0,
    )

    output = child.stdout.readline().strip()
    assert output == "Ready!", output

    os.environ["HADOOP_CONF_DIR"] = "target/test"

    yield "hdfs://127.0.0.1:9000"

    try:
        child.communicate(input="\n", timeout=30)
    except:  # noqa: E722
        child.kill()


@pytest.fixture
def client(minidfs: str):
    client = Client(minidfs)

    try:
        yield client
    finally:
        statuses = list(client.list_status("/"))
        for status in statuses:
            client.delete(status.path, True)


@pytest.fixture
def async_client(minidfs: str):
    client = AsyncClient(minidfs)

    yield client


@pytest_asyncio.fixture
async def aio_benchmark(benchmark: BenchmarkFixture):
    async def run_async_coroutine(func, *args, **kwargs):
        return await func(*args, **kwargs)

    def _wrapper(func, *args, **kwargs):
        assert asyncio.iscoroutinefunction(func)

        @benchmark
        def _():
            return asyncio.get_event_loop().run_until_complete(
                run_async_coroutine(func, *args, **kwargs)
            )

    return _wrapper



import contextlib
import dataclasses
import io
import os
import re
import stat
import time
from datetime import datetime
from tempfile import TemporaryDirectory
from typing import Callable, Iterator, List, Literal, Optional, Tuple, overload

import pytest

from hdfs_native import Client
from hdfs_native.cli import main as cli_main


def assert_not_exists(client: Client, path: str):
    try:
        client.get_file_info(path)
        pytest.fail(f"Expected file not to exist: {path}")
    except FileNotFoundError:
        pass


@overload
def capture_stdout(func: Callable[[], None], text: Literal[False]) -> bytes: ...


@overload
def capture_stdout(func: Callable[[], None], text: Literal[True] = True) -> str: ...


def capture_stdout(func: Callable[[], None], text: bool = True):
    buf = io.BytesIO()
    with contextlib.redirect_stdout(io.TextIOWrapper(buf)) as wrapper:
        func()
        if text:
            wrapper.seek(0)
            return wrapper.read()
        else:
            return buf.getvalue()


def test_cat(client: Client):
    with client.create("/testfile") as file:
        file.write(b"1234")

    output = capture_stdout(lambda: cli_main(["cat", "/testfile"]), False)
    assert output == b"1234"

    with client.create("/testfile2") as file:
        file.write(b"5678")

    output = capture_stdout(lambda: cli_main(["cat", "/testfile", "/testfile2"]), False)
    assert output == b"12345678"

    with pytest.raises(FileNotFoundError):
        cli_main(["cat", "/nonexistent"])


def test_chmod(client: Client):
    with pytest.raises(FileNotFoundError):
        cli_main(["chmod", "755", "/testfile"])

    client.create("/testfile").close()

    cli_main(["chmod", "700", "/testfile"])
    assert client.get_file_info("/testfile").permission == 0o700

    cli_main(["chmod", "007", "/testfile"])
    assert client.get_file_info("/testfile").permission == 0o007

    cli_main(["chmod", "1777", "/testfile"])
    assert client.get_file_info("/testfile").permission == 0o1777

    with pytest.raises(ValueError):
        cli_main(["chmod", "2777", "/testfile"])

    with pytest.raises(ValueError):
        cli_main(["chmod", "2778", "/testfile"])

    client.mkdirs("/testdir")
    client.create("/testdir/testfile").close()
    original_permission = client.get_file_info("/testdir/testfile").permission

    cli_main(["chmod", "700", "/testdir"])
    assert client.get_file_info("/testdir").permission == 0o700
    assert client.get_file_info("/testdir/testfile").permission == original_permission

    cli_main(["chmod", "-R", "700", "/testdir"])
    assert client.get_file_info("/testdir").permission == 0o700
    assert client.get_file_info("/testdir/testfile").permission == 0o700


def test_chown(client: Client):
    with pytest.raises(FileNotFoundError):
        cli_main(["chown", "testuser", "/testfile"])

    client.create("/testfile").close()
    status = client.get_file_info("/testfile")
    group = status.group

    cli_main(["chown", "testuser", "/testfile"])
    status = client.get_file_info("/testfile")
    assert status.owner == "testuser"
    assert status.group == group

    cli_main(["chown", ":testgroup", "/testfile"])
    status = client.get_file_info("/testfile")
    assert status.owner == "testuser"
    assert status.group == "testgroup"

    cli_main(["chown", "newuser:newgroup", "/testfile"])
    status = client.get_file_info("/testfile")
    assert status.owner == "newuser"
    assert status.group == "newgroup"

    client.mkdirs("/testdir")
    client.create("/testdir/testfile").close()
    file_status = client.get_file_info("/testdir/testfile")

    cli_main(["chown", "testuser:testgroup", "/testdir"])
    status = client.get_file_info("/testdir")
    assert status.owner == "testuser"
    assert status.group == "testgroup"
    status = client.get_file_info("/testdir/testfile")
    assert status.owner == file_status.owner
    assert status.group == file_status.group

    cli_main(["chown", "-R", "testuser:testgroup", "/testdir"])
    status = client.get_file_info("/testdir/testfile")
    assert status.owner == "testuser"
    assert status.group == "testgroup"


def test_du(client: Client):
    with client.create("/testfile") as file:
        file.write(b"1234")

    client.mkdirs("/testdir")

    with client.create("/testdir/testfile") as file:
        for i in range(1024):
            file.write(i.to_bytes(4, "big"))

    assert capture_stdout(lambda: cli_main(["du", "/"])).strip().split("\n") == [
        "4096  12288  /testdir ",
        "   4     12  /testfile",
    ]

    assert capture_stdout(lambda: cli_main(["du", "-h", "/"])).strip().split("\n") == [
        "4.0K  12.0K  /testdir ",
        "   4     12  /testfile",
    ]

    assert capture_stdout(lambda: cli_main(["du", "-s", "/"])).strip().split("\n") == [
        "4100  12300  /",
    ]

    assert capture_stdout(lambda: cli_main(["du", "-fh", "/"])).strip().split("\n") == [
        "4.0K  12.0K  /testdir   1  1",
        "   4     12  /testfile  1  0",
    ]

    assert capture_stdout(lambda: cli_main(["du", "-vfh", "/"])).strip().split(
        "\n"
    ) == [
        "File Size  Disk Size  Path       File Count  Directory Count",
        "     4.0K      12.0K  /testdir            1                1",
        "        4         12  /testfile           1                0",
    ]

    assert capture_stdout(
        lambda: cli_main(["du", "-vsfh", "/testdir", "/testfile"])
    ).strip().split("\n") == [
        "File Size  Disk Size  Path       File Count  Directory Count",
        "     4.0K      12.0K  /testdir            1                1",
        "        4         12  /testfile           1                0",
    ]


def test_get(client: Client, monkeypatch: pytest.MonkeyPatch):
    data = b"0123456789"

    with pytest.raises(FileNotFoundError):
        cli_main(["get", "/testfile", "testfile"])

    with client.create("/testfile") as file:
        file.write(data)

    status = client.get_file_info("/testfile")

    with TemporaryDirectory() as tmp_dir:
        cli_main(["get", "/testfile", os.path.join(tmp_dir, "localfile")])
        with open(os.path.join(tmp_dir, "localfile"), "rb") as file:
            assert file.read() == data

        cli_main(["get", "/testfile", tmp_dir])
        with open(os.path.join(tmp_dir, "testfile"), "rb") as file:
            assert file.read() == data

        os.remove(os.path.join(tmp_dir, "testfile"))

        with monkeypatch.context() as m:
            m.chdir(tmp_dir)
            cli_main(["get", "/testfile"])

        with open(os.path.join(tmp_dir, "testfile"), "rb") as file:
            assert file.read() == data

        with pytest.raises(FileExistsError):
            cli_main(["get", "/testfile", tmp_dir])

        cli_main(["get", "-f", "-p", "/testfile", tmp_dir])
        st = os.stat(os.path.join(tmp_dir, "testfile"))
        assert stat.S_IMODE(st.st_mode) == status.permission
        assert int(st.st_atime * 1000) == status.access_time
        assert int(st.st_mtime * 1000) == status.modification_time

    with client.create("/testfile2") as file:
        file.write(data)

    with pytest.raises(ValueError):
        cli_main(["get", "/testfile", "/testfile2", "notadir"])

    with TemporaryDirectory() as tmp_dir:
        cli_main(["get", "/testfile", "/testfile2", tmp_dir])

        with open(os.path.join(tmp_dir, "testfile"), "rb") as file:
            assert file.read() == data

        with open(os.path.join(tmp_dir, "testfile2"), "rb") as file:
            assert file.read() == data


def test_ls(client: Client):
    @dataclasses.dataclass
    class FileOutput:
        permission: str
        replication: str
        size: str
        path: str

    def parse_output(output: str) -> Iterator[Tuple[int, List[FileOutput]]]:
        current_items: Optional[int] = None
        current_batch: List[FileOutput] = []

        for line in output.split("\n"):
            if match := re.match(r"Found (\d)+ items", line):
                if current_items is not None:
                    yield (current_items, current_batch)

                current_items = int(match.group(1))
                current_batch = []

            elif line.strip():
                match = re.match(
                    r"(\S+)\s+(\S+)\s+\S+\s+\S+\s+([0-9.]+\w?)\s+\S+\s+\S+\s+(\S+)",
                    line,
                )
                assert match is not None
                current_batch.append(
                    FileOutput(
                        permission=match.group(1),
                        replication=match.group(2),
                        size=match.group(3),
                        path=match.group(4),
                    )
                )

        if current_items is not None and len(current_batch) > 0:
            yield (current_items, current_batch)

    with pytest.raises(FileNotFoundError):
        cli_main(["ls", "/fake"])

    with client.create("/testfile1") as f:
        f.write(bytes(range(10)))

    # Make sure we wait a few milliseconds so we don't get the exact same timestamp
    time.sleep(0.01)

    with client.create("/testfile2") as f:
        for i in range(1024):
            f.write(i.to_bytes(4, "big"))

    time.sleep(0.01)

    client.mkdirs("/testdir")

    directory = FileOutput("drwxr-xr-x", "-", "0", "/testdir")
    file1 = FileOutput("-rw-r--r--", "3", "10", "/testfile1")
    file2 = FileOutput("-rw-r--r--", "3", "4096", "/testfile2")

    def check_output(command: List[str], expected: List[FileOutput]):
        groups = list(parse_output(capture_stdout(lambda: cli_main(command))))
        assert len(groups) == 1
        assert groups[0][0] == 3
        assert len(groups[0][1]) == 3
        assert groups[0][1] == expected

    check_output(["ls", "/"], [directory, file1, file2])
    check_output(["ls", "-t", "/"], [directory, file2, file1])
    check_output(["ls", "-r", "-t", "/"], [file1, file2, directory])
    check_output(["ls", "-S", "/"], [file2, file1, directory])
    check_output(["ls", "-r", "-S", "/"], [directory, file1, file2])

    check_output(
        ["ls", "-h", "/"], [directory, file1, dataclasses.replace(file2, size="4.0K")]
    )

    output = capture_stdout(lambda: cli_main(["ls", "-C", "/"])).strip().split("\n")
    assert output == [directory.path, file1.path, file2.path]


def test_mkdir(client: Client):
    cli_main(["mkdir", "/testdir"])
    assert client.get_file_info("/testdir").isdir

    with pytest.raises(FileNotFoundError):
        cli_main(["mkdir", "/testdir/nested/dir"])

    cli_main(["mkdir", "-p", "/testdir/nested/dir"])
    assert client.get_file_info("/testdir/nested/dir").isdir


def test_mv(client: Client):
    client.create("/testfile").close()
    client.mkdirs("/testdir")

    cli_main(["mv", "/testfile", "/testfile2"])

    client.get_file_info("/testfile2")

    with pytest.raises(ValueError):
        cli_main(["mv", "/testfile2", "hdfs://badnameservice/testfile"])

    with pytest.raises(FileNotFoundError):
        cli_main(["mv", "/testfile2", "/nonexistent/testfile"])

    cli_main(["mv", "/testfile2", "/testdir"])

    client.get_file_info("/testdir/testfile2")

    client.rename("/testdir/testfile2", "/testfile1")
    client.create("/testfile2").close()

    with pytest.raises(ValueError):
        cli_main(["mv", "/testfile1", "/testfile2", "/testfile3"])

    cli_main(["mv", "/testfile1", "/testfile2", "/testdir/"])

    client.get_file_info("/testdir/testfile1")
    client.get_file_info("/testdir/testfile2")


def test_put(client: Client):
    data = b"0123456789"

    with pytest.raises(FileNotFoundError):
        cli_main(["put", "testfile", "/testfile"])

    with TemporaryDirectory() as tmp_dir:
        with open(os.path.join(tmp_dir, "testfile"), "wb") as file:
            file.write(data)

        cli_main(["put", os.path.join(tmp_dir, "testfile"), "/remotefile"])
        with client.read("/remotefile") as file:
            assert file.read() == data

        cli_main(["put", os.path.join(tmp_dir, "testfile"), "/"])
        with client.read("/testfile") as file:
            assert file.read() == data

        with pytest.raises(FileExistsError):
            cli_main(["put", os.path.join(tmp_dir, "testfile"), "/"])

        cli_main(["put", "-f", "-p", os.path.join(tmp_dir, "testfile"), "/"])
        st = os.stat(os.path.join(tmp_dir, "testfile"))
        status = client.get_file_info("/testfile")
        assert stat.S_IMODE(st.st_mode) == status.permission
        assert int(st.st_atime * 1000) == status.access_time
        assert int(st.st_mtime * 1000) == status.modification_time

        with open(os.path.join(tmp_dir, "testfile2"), "wb") as file:
            file.write(data)

        with pytest.raises(ValueError):
            cli_main(
                [
                    "put",
                    os.path.join(tmp_dir, "testfile"),
                    os.path.join(tmp_dir, "testfile2"),
                    "/notadir",
                ]
            )

        client.mkdirs("/testdir")
        cli_main(
            [
                "put",
                os.path.join(tmp_dir, "testfile"),
                os.path.join(tmp_dir, "testfile2"),
                "/testdir",
            ]
        )

        with client.read("/testdir/testfile") as file:
            assert file.read() == data
        with client.read("/testdir/testfile2") as file:
            assert file.read() == data


def test_rm(client: Client):
    with pytest.raises(ValueError):
        cli_main(["rm", "/testfile"])

    with pytest.raises(FileNotFoundError):
        cli_main(["rm", "-s", "/testfile"])

    cli_main(["rm", "-f", "-s", "/testfile"])

    client.create("/testfile").close()
    cli_main(["rm", "-s", "/testfile"])
    assert_not_exists(client, "/testfile")

    client.mkdirs("/testdir")
    client.create("/testdir/testfile").close()
    client.create("/testdir/testfile2").close()

    with pytest.raises(RuntimeError):
        cli_main(["rm", "-s", "/testdir"])

    cli_main(["rm", "-r", "-s", "/testdir"])
    assert_not_exists(client, "/testdir")


def test_rmdir(client: Client):
    with pytest.raises(FileNotFoundError):
        cli_main(["rmdir", "/testdir"])

    client.mkdirs("/testdir")
    client.create("/testdir/testfile").close()

    with pytest.raises(RuntimeError):
        cli_main(["rmdir", "/testdir"])

    client.delete("/testdir/testfile")

    cli_main(["rmdir", "/testdir"])

    try:
        client.get_file_info("/testdir")
        pytest.fail("Directory was not removed")
    except FileNotFoundError:
        pass


def test_touch(client: Client):
    cli_main(["touch", "/testfile"])
    client.get_file_info("/testfile")

    cli_main(["touch", "-c", "/testfile2"])
    try:
        client.get_file_info("/testfile2")
        pytest.fail("File should not have been created")
    except FileNotFoundError:
        pass

    cli_main(["touch", "-a", "/testfile"])
    status = client.get_file_info("/testfile")
    assert status.access_time > status.modification_time

    cli_main(["touch", "-m", "/testfile"])
    status = client.get_file_info("/testfile")
    assert status.modification_time > status.access_time

    cli_main(["touch", "-t", "20240101:000000", "/testfile"])
    timestamp = int(
        datetime.strptime("20240101:000000", r"%Y%m%d:%H%M%S").timestamp() * 1000
    )
    status = client.get_file_info("/testfile")
    assert status.modification_time == timestamp
    assert status.access_time == timestamp



import io
import secrets
import shutil
import time
import urllib.parse
from contextlib import suppress
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from fsspec import AbstractFileSystem
from fsspec.utils import tokenize

from . import Client, WriteOptions

if TYPE_CHECKING:
    from . import FileStatus

__all__ = ["HdfsFileSystem", "ViewfsFileSystem"]


class BaseFileSystem(AbstractFileSystem):
    root_marker = "/"

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        *args,
        **storage_options,
    ):
        super().__init__(host, port, *args, **storage_options)
        self.host = host
        self.port = port
        url = f"{self.protocol}://"
        if host:
            url += host
            if port:
                url += f":{port}"
        self.client = Client(url, storage_options)

    @property
    def fsid(self):
        return f"hdfs_native_{tokenize(self.protocol, self.host, self.port)}"

    @classmethod
    def _strip_protocol(cls, path: str) -> str:
        url = urllib.parse.urlparse(path)
        return url.path or cls.root_marker

    def unstrip_protocol(self, name: str) -> str:
        path = self._strip_protocol(name)

        url = f"{self.protocol}://"
        if self.host:
            url += self.host
            if self.port:
                url += f":{self.port}"

        return f"{url}{path}"

    @staticmethod
    def _get_kwargs_from_urls(path):
        url = urllib.parse.urlparse(path)
        return {"host": url.hostname, "port": url.port}

    def _convert_file_status(self, file_status: "FileStatus") -> Dict:
        return {
            "name": file_status.path,
            "size": file_status.length,
            "type": "directory" if file_status.isdir else "file",
            "permission": file_status.permission,
            "owner": file_status.owner,
            "group": file_status.group,
            "modification_time": file_status.modification_time,
            "access_time": file_status.access_time,
        }

    def info(self, path, **_kwargs) -> Dict:
        file_status = self.client.get_file_info(self._strip_protocol(path))
        return self._convert_file_status(file_status)

    def exists(self, path, **_kwargs):
        try:
            self.info(path)
            return True
        except FileNotFoundError:
            return False

    def ls(self, path: str, detail=True, **kwargs) -> List[Union[str, Dict]]:
        listing = self.client.list_status(self._strip_protocol(path), False)
        if detail:
            return [self._convert_file_status(status) for status in listing]
        else:
            return [status.path for status in listing]

    def touch(self, path: str, truncate=True, **kwargs):
        path = self._strip_protocol(path)
        if truncate or not self.exists(path):
            with self.open(path, "wb", **kwargs):
                pass
        else:
            now = int(time.time() * 1000)
            self.client.set_times(path, now, now)

    def du(
        self,
        path: str,
        total=True,
        maxdepth: Optional[int] = None,
        withdirs=False,
        **kwargs,
    ) -> Union[int, Dict[str, int]]:
        if total:
            if maxdepth is not None:
                raise NotImplementedError("maxdepth is not supported with total")

            content_summary = self.client.get_content_summary(path)
            return content_summary.length
        else:
            return super().du(path, total, maxdepth, withdirs, **kwargs)

    def mkdir(self, path: str, create_parents=True, **kwargs):
        self.client.mkdirs(
            self._strip_protocol(path),
            kwargs.get("permission", 0o755),
            create_parents,
        )

    def makedirs(self, path: str, exist_ok=False):
        path = self._strip_protocol(path)
        if not exist_ok and self.exists(path):
            raise FileExistsError("File or directory already exists")

        return self.mkdir(path, create_parents=True)

    def mv(self, path1: str, path2: str, **kwargs):
        self.client.rename(
            self._strip_protocol(path1),
            self._strip_protocol(path2),
            kwargs.get("overwrite", False),
        )

    def cp_file(self, path1, path2, **kwargs):
        with self._open(self._strip_protocol(path1), "rb") as lstream:
            tmp_fname = f".{self._strip_protocol(path2)}.tmp.{secrets.token_hex(6)}"
            try:
                with self.open(tmp_fname, "wb") as rstream:
                    shutil.copyfileobj(lstream, rstream)
                self.mv(tmp_fname, path2)
            except BaseException:  # noqa
                with suppress(FileNotFoundError):
                    self.fs.delete_file(tmp_fname)
                raise

    def rmdir(self, path: str) -> None:
        self.client.delete(self._strip_protocol(path), False)

    def rm(self, path: str, recursive=False, maxdepth: Optional[int] = None) -> None:
        if maxdepth is not None:
            raise NotImplementedError("maxdepth is not supported")
        self.client.delete(self._strip_protocol(path), recursive)

    def rm_file(self, path: str):
        self.rm(self._strip_protocol(path))

    def modified(self, path: str):
        file_info = self.client.get_file_info(self._strip_protocol(path))
        return datetime.fromtimestamp(file_info.modification_time / 1000)

    def _open(
        self,
        path: str,
        mode="rb",
        overwrite=True,
        replication: Optional[int] = None,
        block_size: Optional[int] = None,
        **_kwargs,
    ):
        path = self._strip_protocol(path)
        if mode == "rb":
            reader = self.client.read(path)
            if not block_size:
                return reader
            return io.BufferedReader(reader, buffer_size=block_size)
        elif mode == "wb":
            write_options = WriteOptions()
            write_options.overwrite = overwrite
            if replication:
                write_options.replication = replication
            if block_size:
                write_options.block_size = block_size
            return self.client.create(path, write_options=write_options)
        elif mode == "ab":
            return self.client.append(path)
        else:
            raise ValueError(f"Mode {mode} is not supported")


class HdfsFileSystem(BaseFileSystem):
    protocol = "hdfs"


class ViewfsFileSystem(BaseFileSystem):
    protocol = "viewfs"



import functools
import glob
import os
import re
import shutil
import stat
import sys
from argparse import ArgumentParser, Namespace
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timezone
from typing import Dict, List, Optional, Sequence, Tuple, Union
from urllib.parse import urlparse

from hdfs_native import Client
from hdfs_native._internal import FileStatus, WriteOptions

__all__ = ["main"]


@functools.cache
def _get_client(connection_url: Optional[str] = None):
    return Client(connection_url)


def _prefix_for_url(url: str) -> str:
    parsed = urlparse(url)

    if parsed.scheme:
        prefix = f"{parsed.scheme}://{parsed.hostname}"
        if parsed.port:
            prefix += f":{parsed.port}"
        return prefix

    return ""


def _client_for_url(url: str) -> Client:
    parsed = urlparse(url)

    if parsed.scheme:
        connection_url = f"{parsed.scheme}://{parsed.hostname}"
        if parsed.port:
            connection_url += f":{parsed.port}"
        return _get_client(connection_url)
    elif parsed.hostname or parsed.port:
        raise ValueError(
            f"Cannot provide host or port without scheme: {parsed.hostname}"
        )
    else:
        return _get_client()


def _verify_nameservices_match(url: str, *urls: str) -> None:
    first = urlparse(url)

    for url in urls:
        parsed = urlparse(url)
        if first.scheme != parsed.scheme or first.hostname != parsed.hostname:
            raise ValueError(
                f"Protocol and host must match: {first.scheme}://{first.hostname} != {parsed.scheme}://{parsed.hostname}"
            )


def _path_for_url(url: str) -> str:
    return urlparse(url).path


def _glob_path(client: Client, glob: str) -> List[str]:
    # TODO: Actually implement this, for now just pretend we have multiple results
    return [glob]


def _glob_local_path(glob_pattern: str) -> List[str]:
    return glob.glob(glob_pattern)


def _download_file(
    client: Client,
    remote_src: str,
    local_dst: str,
    force: bool = False,
    preserve: bool = False,
) -> None:
    if not force and os.path.exists(local_dst):
        raise FileExistsError(f"{local_dst} already exists, use --force to overwrite")

    with client.read(remote_src) as remote_file:
        with open(local_dst, "wb") as local_file:
            for chunk in remote_file.read_range_stream(0, len(remote_file)):
                local_file.write(chunk)

    if preserve:
        status = client.get_file_info(remote_src)
        os.utime(
            local_dst,
            (status.access_time / 1000, status.modification_time / 1000),
        )
        os.chmod(local_dst, status.permission)


def _upload_file(
    client: Client,
    local_src: str,
    remote_dst: str,
    direct: bool = False,
    force: bool = False,
    preserve: bool = False,
) -> None:
    if not direct and not force:
        # Check if file already exists before we write it to a temporary file
        try:
            client.get_file_info(remote_dst)
            raise FileExistsError(
                f"{remote_dst} already exists, use --force to overwrite"
            )
        except FileNotFoundError:
            pass

    if direct:
        write_destination = remote_dst
    else:
        write_destination = f"{remote_dst}.__COPYING__"

    with open(local_src, "rb") as local_file:
        with client.create(
            write_destination,
            WriteOptions(overwrite=force),
        ) as remote_file:
            shutil.copyfileobj(local_file, remote_file)

    if preserve:
        st = os.stat(local_src)
        client.set_times(
            write_destination,
            int(st.st_mtime * 1000),
            int(st.st_atime * 1000),
        )
        client.set_permission(write_destination, stat.S_IMODE(st.st_mode))

    if not direct:
        client.rename(write_destination, remote_dst, overwrite=force)


def _get_widths(parsed: list[dict]) -> dict[str, int]:
    widths: dict[str, int] = defaultdict(lambda: 0)

    for file in parsed:
        for key, value in file.items():
            if isinstance(value, str):
                widths[key] = max(widths[key], len(value))

    return widths


def _human_size(num: int):
    if num < 1024:
        return str(num)

    adjusted = num / 1024.0
    for unit in ("K", "M", "G", "T", "P", "E", "Z"):
        if abs(adjusted) < 1024.0:
            return f"{adjusted:.1f}{unit}"
        adjusted /= 1024.0
    return f"{adjusted:.1f}Y"


def cat(args: Namespace):
    for src in args.src:
        client = _client_for_url(src)
        for path in _glob_path(client, _path_for_url(src)):
            with client.read(path) as file:
                while chunk := file.read(1024 * 1024):
                    sys.stdout.buffer.write(chunk)

    sys.stdout.buffer.flush()


def chmod(args: Namespace):
    if not re.fullmatch(r"1?[0-7]{3}", args.octalmode):
        raise ValueError(f"Invalid mode supplied: {args.octalmode}")

    permission = int(args.octalmode, base=8)

    for url in args.path:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            if args.recursive:
                for status in client.list_status(path, True):
                    client.set_permission(status.path, permission)
            else:
                client.set_permission(path, permission)


def chown(args: Namespace):
    split = args.owner.split(":")
    if len(split) > 2:
        raise ValueError(f"Invalid owner and group pattern: {args.owner}")

    if len(split) == 1:
        owner = split[0]
        group = None
    else:
        owner = split[0] if split[0] else None
        group = split[1]

    for url in args.path:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            if args.recursive:
                for status in client.list_status(path, True):
                    client.set_owner(status.path, owner, group)
            else:
                client.set_owner(path, owner, group)


def du(args: Namespace):
    parsed: List[Dict[str, str]] = []

    if args.verbose:
        header = {
            "file_size": "File Size",
            "disk_size": "Disk Size",
            "path": "Path",
        }
        if args.file_count:
            header["file_count"] = "File Count"
            header["directory_count"] = "Directory Count"

        parsed.append(header)

    for url in args.path:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            prefix = _prefix_for_url(url)

            if args.summary:
                summaries = [(prefix + path, client.get_content_summary(path))]
            else:
                summaries = []
                for status in client.list_status(path):
                    summaries.append(
                        (prefix + status.path, client.get_content_summary(status.path))
                    )

            for path, summary in summaries:
                if args.human_readable:
                    file_size = _human_size(summary.length)
                    disk_size = _human_size(summary.space_consumed)
                else:
                    file_size = str(summary.length)
                    disk_size = str(summary.space_consumed)

                parsed_file = {
                    "file_size": file_size,
                    "disk_size": disk_size,
                    "path": path,
                }

                if args.file_count:
                    parsed_file["file_count"] = str(summary.file_count)
                    parsed_file["directory_count"] = str(summary.directory_count)

                parsed.append(parsed_file)

    widths = _get_widths(parsed)

    def format(
        file: Dict[str, str],
        field: str,
        right_align: bool = False,
    ):
        value = str(file[field])

        width = len(value)
        if widths and field in widths:
            width = widths[field]

        if right_align:
            return f"{value:>{width}}"
        return f"{value:{width}}"

    for file in parsed:
        formatted_fields = [
            format(file, "file_size", True),
            format(file, "disk_size", True),
            format(file, "path"),
        ]
        if args.file_count:
            formatted_fields.append(format(file, "file_count", True))
            formatted_fields.append(format(file, "directory_count", True))

        print("  ".join(formatted_fields))


def get(args: Namespace):
    paths: List[Tuple[Client, str]] = []

    if len(args.src) > 1:
        srcs = args.src[:-1]
        dst = args.src[-1]
    else:
        srcs = args.src
        dst = os.getcwd()

    for url in srcs:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            paths.append((client, path))

    dst_is_dir = os.path.isdir(dst)

    if len(paths) > 1 and not dst_is_dir:
        raise ValueError("Destination must be directory when copying multiple files")
    elif not dst_is_dir:
        _download_file(
            paths[0][0],
            paths[0][1],
            dst,
            force=args.force,
            preserve=args.preserve,
        )
    else:
        with ThreadPoolExecutor(args.threads) as executor:
            futures = []
            for client, path in paths:
                filename = os.path.basename(path)

                futures.append(
                    executor.submit(
                        _download_file,
                        client,
                        path,
                        os.path.join(dst, filename),
                        force=args.force,
                        preserve=args.preserve,
                    )
                )

            # Iterate to raise any exceptions thrown
            for f in as_completed(futures):
                f.result()


def ls(args: Namespace):
    def parse_status(status: FileStatus, prefix: str) -> Dict[str, Union[int, str]]:
        file_time = status.modification_time
        if args.access_time:
            file_time = status.access_time

        file_time_string = datetime.fromtimestamp(file_time / 1000).strftime(
            r"%Y-%m-%d %H:%M"
        )

        permission = status.permission
        if status.isdir:
            permission |= stat.S_IFDIR
        else:
            permission |= stat.S_IFREG

        mode = stat.filemode(permission)

        if args.human_readable:
            length_string = _human_size(status.length)
        else:
            length_string = str(status.length)

        path = prefix + status.path

        return {
            "mode": mode,
            "replication": str(status.replication) if status.replication else "-",
            "owner": status.owner,
            "group": status.group,
            "length": status.length,
            "length_formatted": length_string,
            "time": file_time,
            "time_formatted": file_time_string,
            "path": path,
        }

    def print_files(
        parsed: List[Dict[str, Union[int, str]]],
        widths: Optional[Dict[str, int]] = None,
    ):
        if args.sort_time:
            parsed = sorted(parsed, key=lambda x: x["time"], reverse=not args.reverse)
        elif args.sort_size:
            parsed = sorted(parsed, key=lambda x: x["length"], reverse=not args.reverse)

        def format(
            file: Dict[str, Union[int, str]],
            field: str,
            right_align: bool = False,
        ):
            value = str(file[field])

            width = len(value)
            if widths and field in widths:
                width = widths[field]

            if right_align:
                return f"{value:>{width}}"
            return f"{value:{width}}"

        for file in parsed:
            if args.path_only:
                print(file["path"])
            else:
                formatted_fields = [
                    format(file, "mode"),
                    format(file, "replication"),
                    format(file, "owner"),
                    format(file, "group"),
                    format(file, "length_formatted", True),
                    format(file, "time_formatted"),
                    format(file, "path"),
                ]
                print(" ".join(formatted_fields))

    for url in args.path:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            status = client.get_file_info(path)

            prefix = _prefix_for_url(url)

            if status.isdir:
                parsed = [
                    parse_status(status, prefix)
                    for status in client.list_status(path, args.recursive)
                ]

                if not args.path_only:
                    print(f"Found {len(parsed)} items")

                widths = _get_widths(parsed)
                print_files(parsed, widths)
            else:
                print_files([parse_status(status, prefix)])


def mkdir(args: Namespace):
    create_parent = args.parent

    for url in args.path:
        client = _client_for_url(url)
        client.mkdirs(_path_for_url(url), create_parent=create_parent)


def mv(args: Namespace):
    _verify_nameservices_match(args.dst, *args.src)

    client = _client_for_url(args.dst)
    dst_path = _path_for_url(args.dst)

    dst_isdir = False
    try:
        dst_isdir = client.get_file_info(dst_path).isdir
    except FileNotFoundError:
        pass

    resolved_src = [
        path
        for pattern in args.src
        for path in _glob_path(client, _path_for_url(pattern))
    ]

    if len(resolved_src) > 1 and not dst_isdir:
        raise ValueError(
            "destination must be a directory if multiple sources are provided"
        )

    for src in resolved_src:
        src_path = _path_for_url(src)
        if dst_isdir:
            target_path = os.path.join(dst_path, os.path.basename(src_path))
        else:
            target_path = dst_path

        client.rename(src_path, target_path)


def put(args: Namespace):
    paths: List[str] = []

    for pattern in args.localsrc:
        for path in _glob_local_path(pattern):
            paths.append(path)

    if len(paths) == 0:
        raise FileNotFoundError("No files matched patterns")

    client = _client_for_url(args.dst)
    dst_path = _path_for_url(args.dst)

    dst_is_dir = False
    try:
        dst_is_dir = client.get_file_info(dst_path).isdir
    except FileNotFoundError:
        pass

    if len(paths) > 1 and not dst_is_dir:
        raise ValueError("Destination must be directory when copying multiple files")
    elif not dst_is_dir:
        _upload_file(
            client,
            paths[0],
            dst_path,
            direct=args.direct,
            force=args.force,
            preserve=args.preserve,
        )
    else:
        with ThreadPoolExecutor(args.threads) as executor:
            futures = []
            for path in paths:
                filename = os.path.basename(path)

                futures.append(
                    executor.submit(
                        _upload_file,
                        client,
                        path,
                        os.path.join(dst_path, filename),
                        direct=args.direct,
                        force=args.force,
                        preserve=args.preserve,
                    )
                )

            # Iterate to raise any exceptions thrown
            for f in as_completed(futures):
                f.result()


def rm(args: Namespace):
    if not args.skip_trash:
        raise ValueError(
            "Moving files to the trash is not currently supported. Pass --skip-trash to permanently delete the files."
        )

    for url in args.src:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            if not client.delete(path, args.recursive) and not args.force:
                raise FileNotFoundError(f"Failed to delete {path}")


def rmdir(args: Namespace):
    for url in args.dir:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            status = client.get_file_info(path)
            if not status.isdir:
                raise ValueError(f"{path} is not a directory")

            client.delete(path)


def touch(args: Namespace):
    for url in args.path:
        client = _client_for_url(url)
        for path in _glob_path(client, _path_for_url(url)):
            timestamp = None
            if args.timestamp:
                timestamp = datetime.strptime(args.timestamp, r"%Y%m%d:%H%M%S")

            try:
                status = client.get_file_info(path)
                if timestamp is None:
                    timestamp = datetime.now(timezone.utc)
            except FileNotFoundError:
                if args.only_change:
                    return

                client.create(path).close()
                status = client.get_file_info(path)

            if timestamp:
                access_time = int(timestamp.timestamp() * 1000)
                modification_time = int(timestamp.timestamp() * 1000)
                if args.access_time:
                    modification_time = status.modification_time
                if args.modification_time:
                    access_time = status.access_time

                client.set_times(path, modification_time, access_time)


def main(in_args: Optional[Sequence[str]] = None):
    parser = ArgumentParser(
        description="""Command line utility for interacting with HDFS using hdfs-native.
        Globs are not currently supported, all file paths are treated as exact paths."""
    )

    subparsers = parser.add_subparsers(title="Subcommands", required=True)

    cat_parser = subparsers.add_parser(
        "cat",
        help="Print the contents of a file",
        description="Print the contents of a file to stdout",
        add_help=False,
    )
    cat_parser.add_argument("src", nargs="+", help="File pattern to print")
    cat_parser.set_defaults(func=cat)

    chmod_parser = subparsers.add_parser(
        "chmod",
        help="Changes permissions of a file",
        description="Changes permissions of a file. Only octal permissions are supported.",
        add_help=False,
    )
    chmod_parser.add_argument(
        "-R",
        "--recursive",
        action="store_true",
        help="Modify files recursively",
    )
    chmod_parser.add_argument(
        "octalmode",
        help="The mode to set the permission to in octal format",
    )
    chmod_parser.add_argument("path", nargs="+", help="File pattern to update")
    chmod_parser.set_defaults(func=chmod)

    chown_parser = subparsers.add_parser(
        "chown",
        help="Changes owner and group of a file",
        description="Changes owner and group of a file",
        add_help=False,
    )
    chown_parser.add_argument(
        "-R",
        "--recursive",
        action="store_true",
        help="Modify files recursively",
    )
    chown_parser.add_argument(
        "owner",
        metavar="[owner[:group] | :group]",
        help="Owner and group to set for the file",
    )
    chown_parser.add_argument("path", nargs="+", help="File pattern to modify")
    chown_parser.set_defaults(func=chown)

    du_parser = subparsers.add_parser(
        "du",
        help="Show the amount of space used by the files that match the specified file pattern",
        description="Show the amount of space used by the files that match the specified file pattern",
        add_help=False,
    )
    du_parser.add_argument(
        "-s",
        "--summary",
        action="store_true",
        default=False,
        help="Show the total size of matching directories instead of traversing their children",
    )
    du_parser.add_argument(
        "-h",
        "--human-readable",
        action="store_true",
        default=False,
        help="Format the size in a human-readable fashion",
    )
    du_parser.add_argument(
        "-f",
        "--file-count",
        action="store_true",
        default=False,
        help="Include the file count",
    )
    du_parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Include a header line",
    )
    du_parser.add_argument("path", nargs="+")
    du_parser.set_defaults(func=du)

    get_parser = subparsers.add_parser(
        "get",
        aliases=["copyToLocal"],
        help="Copy files to a local destination",
        description="""Copy files matching a pattern to a local destination.
            When copying multiple files, the destination must be a directory""",
        add_help=False,
    )
    get_parser.add_argument(
        "-p",
        "--preserve",
        action="store_true",
        default=False,
        help="Preserve timestamps and the mode",
    )
    get_parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Overwrite the destination if it already exists",
    )
    get_parser.add_argument(
        "-t",
        "--threads",
        type=int,
        help="Number of threads to use",
        default=1,
    )
    get_parser.add_argument(
        "src",
        nargs="+",
        help="Source patterns to copy",
    )
    get_parser.add_argument(
        "localdst",
        nargs="?",
        help="Local destination to write to. Defaults to current directory.",
    )
    get_parser.set_defaults(func=get)

    ls_parser = subparsers.add_parser(
        "ls",
        help="List contents that match the specified patterns",
        description="""List contents that match the specified patterns. For a directory, list its
        direct children.""",
        add_help=False,
    )
    ls_parser.add_argument(
        "-C",
        "--path-only",
        action="store_true",
        default=False,
        help="Display the path of files and directories only.",
    )
    ls_parser.add_argument(
        "-h",
        "--human-readable",
        action="store_true",
        default=False,
        help="Formats the sizes of files in a human-readable fashion rather than a number of bytes",
    )
    ls_parser.add_argument(
        "-R",
        "--recursive",
        action="store_true",
        default=False,
        help="Recursively list the contents of directories",
    )
    ls_parser.add_argument(
        "-t",
        "--sort-time",
        action="store_true",
        default=False,
        help="Sort files by modification time (most recent first)",
    )
    ls_parser.add_argument(
        "-S",
        "--sort-size",
        action="store_true",
        default=False,
        help="Sort files by size (largest first)",
    )
    ls_parser.add_argument(
        "-r",
        "--reverse",
        action="store_true",
        default=False,
        help="Reverse the order of the sort",
    )
    ls_parser.add_argument(
        "-u",
        "--access-time",
        action="store_true",
        default=False,
        help="Use the last access time instead of modification time for display and sorting",
    )
    ls_parser.add_argument("path", nargs="+", help="Path to display contents of")
    ls_parser.set_defaults(func=ls)

    mkdir_parser = subparsers.add_parser(
        "mkdir",
        help="Create a directory",
        description="Create a directory in a specified path",
        add_help=False,
    )
    mkdir_parser.add_argument(
        "path",
        nargs="+",
        help="Path for the directory to create",
    )
    mkdir_parser.add_argument(
        "-p",
        "--parent",
        action="store_true",
        help="Create any missing parent directories",
    )
    mkdir_parser.set_defaults(func=mkdir)

    mv_parser = subparsers.add_parser(
        "mv",
        help="Move files or directories",
        description="""Move a file or directory from <src> to <dst>. Must be part of the same name service.
        If multiple src are provided, dst must be a directory""",
        add_help=False,
    )
    mv_parser.add_argument("src", nargs="+", help="Files or directories to move")
    mv_parser.add_argument("dst", help="Target destination of file or directory")
    mv_parser.set_defaults(func=mv)

    put_parser = subparsers.add_parser(
        "put",
        aliases=["copyFromLocal"],
        help="Copy local files to a remote destination",
        description="""Copy files matching a pattern to a remote destination.
            When copying multiple files, the destination must be a directory""",
        add_help=False,
    )
    put_parser.add_argument(
        "-p",
        "--preserve",
        action="store_true",
        default=False,
        help="Preserve timestamps, ownership, and the mode",
    )
    put_parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Overwrite the destination if it already exists",
    )
    put_parser.add_argument(
        "-d",
        "--direct",
        action="store_true",
        default=False,
        help="Skip creation of temporary file (<dst>._COPYING_) and write directly to file",
    )
    put_parser.add_argument(
        "-t",
        "--threads",
        type=int,
        help="Number of threads to use",
        default=1,
    )
    put_parser.add_argument(
        "localsrc",
        nargs="+",
        help="Source patterns to copy",
    )
    put_parser.add_argument(
        "dst",
        help="Local destination to write to",
    )
    put_parser.set_defaults(func=put)

    rm_parser = subparsers.add_parser(
        "rm",
        help="Delete files",
        description="Delete all files matching the specified file patterns",
        add_help=False,
    )
    rm_parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Ignore if the file does not exist",
    )
    rm_parser.add_argument(
        "-r",
        "-R",
        "--recursive",
        action="store_true",
        default=False,
        help="Recursively delete directories",
    )
    rm_parser.add_argument(
        "-s",
        "--skip-trash",
        action="store_true",
        default=False,
        help="Permanently delete files instead of moving them to the trash",
    )
    rm_parser.add_argument(
        "src",
        nargs="+",
        help="File patterns to delete",
    )
    rm_parser.set_defaults(func=rm)

    rmdir_parser = subparsers.add_parser(
        "rmdir",
        help="Delete an empty directory",
        description="Delete an empty directory",
        add_help=False,
    )
    rmdir_parser.add_argument(
        "dir",
        nargs="+",
        help="Source patterns to copy",
    )
    rmdir_parser.set_defaults(func=rmdir)

    touch_parser = subparsers.add_parser(
        "touch",
        help="Updates the access and modification times of a file or creates it if it doesn't exist",
        description="""Updates the access and modification times of the file specified by the <path> to
                    the current time. If the file does not exist, then a zero length file is created
                    at <path> with current time as the timestamp of that <path>.""",
        add_help=False,
    )
    touch_parser_time_group = touch_parser.add_mutually_exclusive_group()
    touch_parser_time_group.add_argument(
        "-a",
        "--access-time",
        action="store_true",
        default=False,
        help="Only change the access time",
    )
    touch_parser_time_group.add_argument(
        "-m",
        "--modification-time",
        action="store_true",
        default=False,
        help="Only change the modification time",
    )
    touch_parser.add_argument(
        "-t",
        "--timestamp",
        help="Use specified timestamp instead of current time in the format yyyyMMdd:HHmmss",
    )
    touch_parser.add_argument(
        "-c",
        "--only-change",
        action="store_true",
        default=False,
        help="Don't create the file if it doesn't exist",
    )
    touch_parser.add_argument(
        "path",
        nargs="+",
    )
    touch_parser.set_defaults(func=touch)

    def show_help(args: Namespace):
        subparsers.choices[args.cmd].print_help()

    subparser_keys = list(subparsers.choices.keys())

    help_parser = subparsers.add_parser(
        "help",
        help="Display usage of a subcommand",
        description="Display usage of a subcommand",
    )
    help_parser.add_argument(
        "cmd",
        choices=subparser_keys,
        help="Command to show usage for",
    )
    help_parser.set_defaults(func=show_help)

    args = parser.parse_args(in_args)
    args.func(args)


if __name__ == "__main__":
    main()



import urllib.parse
from abc import ABC, abstractmethod
from datetime import datetime

import fsspec
import pytest

from hdfs_native.fsspec import BaseFileSystem


def test_config(minidfs: str):
    url = urllib.parse.urlparse(minidfs)
    fs: BaseFileSystem = fsspec.filesystem(url.scheme, **{"fs.defaultFS": minidfs})
    assert len(fs.ls("/")) == 0


class TestFsspecBase(ABC):
    @abstractmethod
    def file_system(self, url: str) -> BaseFileSystem:
        pass

    @pytest.fixture(scope="class")
    def fs(self, minidfs: str) -> BaseFileSystem:
        return self.file_system(minidfs)

    def test_dirs(self, fs: BaseFileSystem):
        fs.mkdir("/testdir")
        assert fs.info("/testdir")["type"] == "directory"

        with pytest.raises(FileExistsError):
            fs.makedirs("/testdir", exist_ok=False)

        fs.makedirs("/testdir", exist_ok=True)

        fs.mkdir("/testdir/nested/dir")
        assert fs.info("/testdir/nested/dir")["type"] == "directory"

        with pytest.raises(FileNotFoundError):
            fs.mkdir("/testdir/nested2/dir", create_parents=False)

        with pytest.raises(RuntimeError):
            fs.rm("/testdir", recursive=False)

        fs.rm("/testdir", recursive=True)

        assert not fs.exists("/testdir")

    def test_io(self, fs: BaseFileSystem):
        with fs.open("/test", mode="wb") as file:
            file.write(b"hello there")

        with fs.open("/test", mode="rb") as file:
            data = file.read()
            assert data == b"hello there"

        with fs.open("/test", mode="rb", block_size=1024) as file:
            data = file.read()
            assert data == b"hello there"

        fs.write_bytes("/test2", b"hello again")
        assert fs.read_bytes("/test2") == b"hello again"
        assert fs.read_bytes("/test2", start=1) == b"ello again"
        assert fs.read_bytes("/test2", end=-1) == b"hello agai"

        fs.mv("/test2", "/test3")
        assert fs.read_text("/test3") == "hello again"
        assert not fs.exists("/test2")

        millis = fs.info("/test3")["modification_time"]
        expected_dt = datetime.fromtimestamp(millis / 1000)
        modified_dt = fs.modified("/test3")
        assert isinstance(modified_dt, datetime)
        assert modified_dt == expected_dt

        fs.rm("/test")
        fs.rm("/test3")

    def test_listing(self, fs: BaseFileSystem):
        fs.mkdir("/testdir")

        fs.touch("/testdir/test1")
        fs.touch("/testdir/test2")

        assert fs.ls("/", detail=False) == ["/testdir"]
        assert fs.ls("/testdir", detail=False) == ["/testdir/test1", "/testdir/test2"]

        listing = fs.ls("/", detail=True)
        assert len(listing) == 1
        assert isinstance(listing[0], dict)
        assert listing[0]["size"] == 0
        assert listing[0]["name"] == "/testdir"
        assert listing[0]["type"] == "directory"

        fs.rm("/testdir", True)

    def test_du(self, fs: BaseFileSystem):
        with fs.open("/test", mode="wb") as file:
            file.write(b"hello there")

        with fs.open("/test2", mode="wb") as file:
            file.write(b"hello again")

        assert fs.du("/test") == 11
        assert fs.du("/test2") == 11
        assert fs.du("/") == 22

        assert fs.du("/", total=False) == {"/test": 11, "/test2": 11}


class TestBaseFileSystem(TestFsspecBase):
    def file_system(self, minidfs: str):
        url = urllib.parse.urlparse(minidfs)

        return fsspec.filesystem("hdfs", host=url.hostname, port=url.port)


class TestViewfsFileSystem(TestFsspecBase):
    def file_system(self, minidfs: str):
        return fsspec.filesystem(
            "viewfs",
            host="test",
            **{"fs.viewfs.mounttable.test.linkFallback": minidfs},
        )


def test_parsing(minidfs: str):
    with fsspec.open(f"{minidfs}/test", "wb") as f:
        f.write(b"hey there")

    url = urllib.parse.urlparse(minidfs)
    fs: BaseFileSystem
    urlpath: str
    fs, urlpath = fsspec.url_to_fs(f"{minidfs}/path")
    assert fs.host == url.hostname
    assert fs.port == url.port
    assert urlpath == "/path"

    assert fs.unstrip_protocol("/path") == f"{minidfs}/path"



import io
import os
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional

# For some reason mypy doesn't think this exists
from typing_extensions import Buffer  # type: ignore

from ._internal import (
    AclEntry,
    AclStatus,
    AsyncRawClient,
    ContentSummary,
    FileStatus,
    RawClient,
    WriteOptions,
)

if TYPE_CHECKING:
    from ._internal import (
        AsyncRawFileReader,
        AsyncRawFileWriter,
        RawFileReader,
        RawFileWriter,
    )

__all__ = [
    "Client",
    "FileReader",
    "FileWriter",
    "AsyncClient",
    "AsyncFileReader",
    "AsyncFileWriter",
    "ContentSummary",
    "FileStatus",
    "WriteOptions",
    "AclEntry",
    "AclStatus",
]


class FileReader(io.RawIOBase):
    def __init__(self, inner: "RawFileReader"):
        self.inner = inner

    def __len__(self) -> int:
        return self.inner.file_length()

    def __iter__(self) -> Iterator[bytes]:
        return self.read_range_stream(0, len(self))

    def __enter__(self):
        # Don't need to do anything special here
        return self

    def __exit__(self, *_args):
        # Future updates could close the file manually here if that would help clean things up
        pass

    @property
    def size(self):
        return len(self)

    def seek(self, offset: int, whence=os.SEEK_SET):
        """Seek to `offset` relative to `whence`"""
        if whence == os.SEEK_SET:
            self.inner.seek(offset)
        elif whence == os.SEEK_CUR:
            self.inner.seek(self.tell() + offset)
        elif whence == os.SEEK_END:
            self.inner.seek(self.inner.file_length() + offset)
        else:
            raise ValueError(f"Unsupported whence {whence}")

    def seekable(self):
        return True

    def tell(self) -> int:
        return self.inner.tell()

    def readable(self) -> bool:
        return True

    def read(self, size: int = -1) -> bytes:
        """Read up to `size` bytes from the file, or all content if -1"""
        return self.inner.read(size)

    def readall(self) -> bytes:
        return self.read()

    def read_range(self, offset: int, len: int) -> bytes:
        """Read `len` bytes from the file starting at `offset`. Doesn't affect the position in the file"""
        return self.inner.read_range(offset, len)

    def read_range_stream(self, offset: int, len: int) -> Iterator[bytes]:
        """
        Read `len` bytes from the file starting at `offset` as an iterator of bytes. Doesn't affect
        the position in the file.

        This is the most efficient way to iteratively read a file.
        """
        return self.inner.read_range_stream(offset, len)

    def close(self) -> None:
        pass


class FileWriter(io.RawIOBase):
    def __init__(self, inner: "RawFileWriter"):
        self.inner = inner

    def writable(self) -> bool:
        return True

    def write(self, buf: Buffer) -> int:
        """Writes `buf` to the file. Always writes all bytes"""
        return self.inner.write(buf)

    def close(self) -> None:
        """Closes the file and saves the final metadata to the NameNode"""
        self.inner.close()

    def __enter__(self) -> "FileWriter":
        return self

    def __exit__(self, *_args):
        self.close()


class Client:
    def __init__(
        self,
        url: Optional[str] = None,
        config: Optional[Dict[str, str]] = None,
    ):
        self.inner = RawClient(url, config)

    def get_file_info(self, path: str) -> FileStatus:
        """Gets the file status for the file at `path`"""
        return self.inner.get_file_info(path)

    def list_status(self, path: str, recursive: bool = False) -> Iterator[FileStatus]:
        """Gets the status of files rooted at `path`. If `recursive` is true, lists all files recursively."""
        return self.inner.list_status(path, recursive)

    def read(self, path: str) -> FileReader:
        """Opens a file for reading at `path`"""
        return FileReader(self.inner.read(path))

    def create(
        self,
        path: str,
        write_options: Optional[WriteOptions] = None,
    ) -> FileWriter:
        """Creates a new file and opens it for writing at `path`"""
        if not write_options:
            write_options = WriteOptions()

        return FileWriter(self.inner.create(path, write_options))

    def append(self, path: str) -> FileWriter:
        """Opens an existing file to append to at `path`"""
        return FileWriter(self.inner.append(path))

    def mkdirs(
        self,
        path: str,
        permission: int = 0o0755,
        create_parent: bool = False,
    ) -> None:
        """
        Creates a directory at `path` with unix permissions `permission`. If `create_parent` is true,
        any parent directories that don't exist will also be created. Otherwise this will fail if
        all parent directories don't already exist.
        """
        return self.inner.mkdirs(path, permission, create_parent)

    def rename(self, src: str, dst: str, overwrite: bool = False) -> None:
        """
        Moves a file or directory from `src` to `dst`. If `overwrite` is True, the destination will be
        overriden if it already exists, otherwise the operation will fail if the destination
        exists.
        """
        return self.inner.rename(src, dst, overwrite)

    def delete(self, path: str, recursive: bool = False) -> bool:
        """
        Deletes a file or directory at `path`. If `recursive` is True and the target is a directory,
        this will delete all contents underneath the directory. If `recursive` is False and the target
        is a non-empty directory, this will fail.
        """
        return self.inner.delete(path, recursive)

    def set_times(self, path: str, mtime: int, atime: int) -> None:
        """
        Changes the modification time and access time of the file at `path` to `mtime` and `atime`, respectively.
        """
        return self.inner.set_times(path, mtime, atime)

    def set_owner(
        self,
        path: str,
        owner: Optional[str] = None,
        group: Optional[str] = None,
    ) -> None:
        """
        Sets the owner and/or group for the file at `path`
        """
        return self.inner.set_owner(path, owner, group)

    def set_permission(self, path: str, permission: int) -> None:
        """
        Sets the permissions for file at `path` to the octal value `permission`.
        For example, to set "rw-r--r--" Unix style permissions, use permission=0o644.
        """
        return self.inner.set_permission(path, permission)

    def set_replication(self, path: str, replication: int) -> bool:
        """
        Sets the replication for file at `path` to `replication`
        """
        return self.inner.set_replication(path, replication)

    def get_content_summary(self, path: str) -> ContentSummary:
        """
        Gets a content summary for `path`
        """
        return self.inner.get_content_summary(path)

    def modify_acl_entries(self, path: str, entries: List[AclEntry]) -> None:
        """
        Update ACL entries for file or directory at `path`. Existing entries will remain.
        """
        return self.inner.modify_acl_entries(path, entries)

    def remove_acl_entries(self, path: str, entries: List[AclEntry]) -> None:
        """
        Remove specific ACL entries for file or directory at `path`.
        """
        return self.inner.remove_acl_entries(path, entries)

    def remove_default_acl(self, path: str) -> None:
        """
        Remove all default ACLs for file or directory at `path`.
        """
        return self.inner.remove_default_acl(path)

    def remove_acl(self, path: str) -> None:
        """
        Remove all ACL entries for file or directory at `path`.
        """
        return self.inner.remove_acl(path)

    def set_acl(self, path: str, entries: List[AclEntry]) -> None:
        """
        Override all ACL entries for file or directory at `path`. If only access ACLs are provided,
        default ACLs are maintained. Likewise if only default ACLs are provided, access ACLs are
        maintained.
        """
        return self.inner.set_acl(path, entries)

    def get_acl_status(self, path: str) -> AclStatus:
        """
        Get the ACL status for the file or directory at `path`.
        """
        return self.inner.get_acl_status(path)


class AsyncFileReader:
    def __init__(self, inner: "AsyncRawFileReader"):
        self.inner = inner

    def __len__(self) -> int:
        return self.inner.file_length()

    def __aiter__(self) -> AsyncIterator[bytes]:
        return self.read_range_stream(0, len(self))

    async def __aenter__(self):
        # Don't need to do anything special here
        return self

    async def __aexit__(self, *_args):
        # Future updates could close the file manually here if that would help clean things up
        pass

    @property
    def size(self):
        return len(self)

    def seek(self, offset: int, whence=os.SEEK_SET):
        """Seek to `offset` relative to `whence`"""
        if whence == os.SEEK_SET:
            self.inner.seek(offset)
        elif whence == os.SEEK_CUR:
            self.inner.seek(self.tell() + offset)
        elif whence == os.SEEK_END:
            self.inner.seek(self.inner.file_length() + offset)
        else:
            raise ValueError(f"Unsupported whence {whence}")

    def seekable(self):
        return True

    def tell(self) -> int:
        return self.inner.tell()

    def readable(self) -> bool:
        return True

    async def read(self, size: int = -1) -> bytes:
        """Read up to `size` bytes from the file, or all content if -1"""
        return await self.inner.read(size)

    async def readall(self) -> bytes:
        return await self.read()

    async def read_range(self, offset: int, len: int) -> bytes:
        """Read `len` bytes from the file starting at `offset`. Doesn't affect the position in the file"""
        return await self.inner.read_range(offset, len)

    def read_range_stream(self, offset: int, len: int) -> AsyncIterator[bytes]:
        """
        Read `len` bytes from the file starting at `offset` as an iterator of bytes. Doesn't affect
        the position in the file.

        This is the most efficient way to iteratively read a file.
        """
        return self.inner.read_range_stream(offset, len)

    async def close(self) -> None:
        pass


class AsyncFileWriter:
    def __init__(self, inner: "AsyncRawFileWriter"):
        self.inner = inner

    def writable(self) -> bool:
        return True

    async def write(self, buf: Buffer) -> int:
        """Writes `buf` to the file. Always writes all bytes"""
        return await self.inner.write(buf)

    async def close(self) -> None:
        """Closes the file and saves the final metadata to the NameNode"""
        await self.inner.close()

    async def __aenter__(self) -> "AsyncFileWriter":
        return self

    async def __aexit__(self, *_args):
        await self.close()


class AsyncClient:
    def __init__(
        self,
        url: Optional[str] = None,
        config: Optional[Dict[str, str]] = None,
    ):
        self.inner = AsyncRawClient(url, config)

    async def get_file_info(self, path: str) -> FileStatus:
        """Gets the file status for the file at `path`"""
        return await self.inner.get_file_info(path)

    def list_status(
        self,
        path: str,
        recursive: bool = False,
    ) -> AsyncIterator[FileStatus]:
        """Gets the status of files rooted at `path`. If `recursive` is true, lists all files recursively."""
        return self.inner.list_status(path, recursive)

    async def read(self, path: str) -> AsyncFileReader:
        """Opens a file for reading at `path`"""
        return AsyncFileReader(await self.inner.read(path))

    async def create(
        self,
        path: str,
        write_options: Optional[WriteOptions] = None,
    ) -> AsyncFileWriter:
        """Creates a new file and opens it for writing at `path`"""
        if not write_options:
            write_options = WriteOptions()

        return AsyncFileWriter(await self.inner.create(path, write_options))

    async def append(self, path: str) -> AsyncFileWriter:
        """Opens an existing file to append to at `path`"""
        return AsyncFileWriter(await self.inner.append(path))

    async def mkdirs(
        self,
        path: str,
        permission: int = 0o0755,
        create_parent: bool = False,
    ) -> None:
        """
        Creates a directory at `path` with unix permissions `permission`. If `create_parent` is true,
        any parent directories that don't exist will also be created. Otherwise this will fail if
        all parent directories don't already exist.
        """
        return await self.inner.mkdirs(path, permission, create_parent)

    async def rename(self, src: str, dst: str, overwrite: bool = False) -> None:
        """
        Moves a file or directory from `src` to `dst`. If `overwrite` is True, the destination will be
        overriden if it already exists, otherwise the operation will fail if the destination
        exists.
        """
        return await self.inner.rename(src, dst, overwrite)

    async def delete(self, path: str, recursive: bool = False) -> bool:
        """
        Deletes a file or directory at `path`. If `recursive` is True and the target is a directory,
        this will delete all contents underneath the directory. If `recursive` is False and the target
        is a non-empty directory, this will fail.
        """
        return await self.inner.delete(path, recursive)

    async def set_times(self, path: str, mtime: int, atime: int) -> None:
        """
        Changes the modification time and access time of the file at `path` to `mtime` and `atime`, respectively.
        """
        return await self.inner.set_times(path, mtime, atime)

    async def set_owner(
        self,
        path: str,
        owner: Optional[str] = None,
        group: Optional[str] = None,
    ) -> None:
        """
        Sets the owner and/or group for the file at `path`
        """
        return await self.inner.set_owner(path, owner, group)

    async def set_permission(self, path: str, permission: int) -> None:
        """
        Sets the permissions for file at `path` to the octal value `permission`.
        For example, to set "rw-r--r--" Unix style permissions, use permission=0o644.
        """
        return await self.inner.set_permission(path, permission)

    async def set_replication(self, path: str, replication: int) -> bool:
        """
        Sets the replication for file at `path` to `replication`
        """
        return await self.inner.set_replication(path, replication)

    async def get_content_summary(self, path: str) -> ContentSummary:
        """
        Gets a content summary for `path`
        """
        return await self.inner.get_content_summary(path)

    async def modify_acl_entries(self, path: str, entries: List[AclEntry]) -> None:
        """
        Update ACL entries for file or directory at `path`. Existing entries will remain.
        """
        return await self.inner.modify_acl_entries(path, entries)

    async def remove_acl_entries(self, path: str, entries: List[AclEntry]) -> None:
        """
        Remove specific ACL entries for file or directory at `path`.
        """
        return await self.inner.remove_acl_entries(path, entries)

    async def remove_default_acl(self, path: str) -> None:
        """
        Remove all default ACLs for file or directory at `path`.
        """
        return await self.inner.remove_default_acl(path)

    async def remove_acl(self, path: str) -> None:
        """
        Remove all ACL entries for file or directory at `path`.
        """
        return await self.inner.remove_acl(path)

    async def set_acl(self, path: str, entries: List[AclEntry]) -> None:
        """
        Override all ACL entries for file or directory at `path`. If only access ACLs are provided,
        default ACLs are maintained. Likewise if only default ACLs are provided, access ACLs are
        maintained.
        """
        return await self.inner.set_acl(path, entries)

    async def get_acl_status(self, path: str) -> AclStatus:
        """
        Get the ACL status for the file or directory at `path`.
        """
        return await self.inner.get_acl_status(path)



import asyncio
from concurrent.futures import ThreadPoolExecutor, as_completed

import pytest
from pytest_benchmark.fixture import BenchmarkFixture

from hdfs_native import AsyncClient, Client


@pytest.mark.benchmark
def test_benchmark_threading(client: Client, benchmark: BenchmarkFixture):
    def do_work():
        def func(path: str):
            client.create(path).close()
            return client.delete(path)

        with ThreadPoolExecutor(10) as executor:
            futures = []
            for i in range(1000):
                futures.append(executor.submit(func, f"/bench{i}"))

            for future in as_completed(futures):
                assert future.result()

    benchmark(do_work)


@pytest.mark.benchmark
def test_benchmark_threading_async(
    async_client: AsyncClient,
    aio_benchmark,
):
    async def do_work():
        async def func(path: str):
            await (await async_client.create(path)).close()
            return await async_client.delete(path)

        async with asyncio.TaskGroup() as tg:
            for i in range(1000):
                tg.create_task(func(f"/bench{i}"))

    aio_benchmark(do_work)
